package denim;

/**
 * Created by Work on 31/10/2016.
 */

import beast.core.BEASTObject;
import beast.core.Description;
import beast.core.Input;
import beast.core.Loggable;
import beast.core.parameter.RealParameter;
import denim.util.InverseGammaMixture;

import java.io.PrintStream;

import static beast.util.Randomizer.nextGamma;
import static beast.util.Randomizer.randomChoicePDF;

/**
 *  Created by Graham Jones on 26/09/2014.
 */

@Description("Statistic that samples the the overall population scale factor multiplied by the per-branch inverse gamma mixture.")


public class PopSampleStatistic extends BEASTObject implements Loggable {
    @SuppressWarnings({"CanBeFinal", "WeakerAccess"})
    public Input<RealParameter> popPriorScaleInput =
            new Input<RealParameter>("popPriorScale",
                    "Overall scale for population size",
                    Input.Validate.REQUIRED);


    @SuppressWarnings({"CanBeFinal", "WeakerAccess"})
    public Input<CoalescenceMigrationDistribution> coalMigDistInput =
            new Input<CoalescenceMigrationDistribution>("coalMigDist",
                    "The CoalescenceMigrationDistribution",
                    Input.Validate.REQUIRED);


    public void initAndValidate() {

    }

    @Override
    public void init(PrintStream out) {
        out.print("PopSize\t");
    }

    @Override
    public void log(int nSample, PrintStream out) {
        out.print(""+getStatisticValue() + "\t");
    }

    @Override
    public void close(PrintStream out) {
        // nothing to do
    }




    private double getStatisticValue() {
        InverseGammaMixture igm = coalMigDistInput.get().getInverseGammaMixture();
        double [] a = igm.getAlphas();
        double [] b = igm.getBetas();
        int i = randomChoicePDF(igm.getWeights());
        double x = nextGamma(a[i], b[i]);
        return popPriorScaleInput.get().getValue() / x;
    }
}
