package denim.dbugtune;

import beast.evolution.tree.Node;
import beast.evolution.tree.Tree;
import beast.evolution.tree.TreeInterface;
import denim.util.BitUnion;
import denim.util.Misc;
import denim.util.UnionArrays;

import java.util.List;

/**
 * Created by Work on 03/11/2016.
 */
public class Checks {



    public static void allTreesAreOK(TreeInterface sTree, List<Tree> gTrees,
                                                String opName, String when) {
        if (!Checks.treeIsOK(sTree)) {
            System.err.println("BUG found in " + opName + ". Bad sTree " + when);
            System.err.println(Misc.allTreesAsText(sTree, gTrees));
            throw new RuntimeException("Fatal STACEY error.");
        }
        for (TreeInterface gTree : gTrees) {
            if (!Checks.treeIsOK(gTree)) {
                System.err.println("BUG found in " + opName + ". Bad gTree " + when);
                System.err.println(Misc.allTreesAsText(sTree, gTrees));
                throw new RuntimeException("Fatal STEPMILS error.");
            }
        }
    }


    public static boolean allTreesAreOK(TreeInterface sTree, List<Tree> gTrees) {
        if (!Checks.treeIsOK(sTree)) {
            return false;
        }
        for (TreeInterface gTree : gTrees) {
            if (!Checks.treeIsOK(gTree)) {
                return false;
            }
        }
        return true;
    }



    public static boolean treeIsOK(TreeInterface tree) {
        Node nodes [] = tree.getNodesAsArray();
        for (int n = 0; n <nodes.length; n++) {
            if (!Double.isFinite(nodes[n].getHeight())) {
                return false;
            }
            if (nodes[n].getHeight() < 0) {
                return false;
            }
        }
        for (int n = 0; n <nodes.length; n++) {
            if (!nodes[n].isLeaf()) {
                if (nodes[n].getChildCount() != 2) {
                    return false;
                }
                double nodeHeight = nodes[n].getHeight();
                if (nodes[n].getLeft().getHeight() > nodeHeight) {
                    return false;
                }
                if (nodes[n].getRight().getHeight() > nodeHeight) {
                    return false;
                }
            }
        }
        return true;
    }



    private static boolean allUnionsLookOK(TreeInterface sTree, List<Tree> gTrees, UnionArrays unionArrays) {
        BitUnion [] sUnions = new BitUnion[sTree.getNodeCount()];
        for (int n = 0; n <sTree.getNodeCount(); n++) {
            sUnions[n] = unionArrays.sNodeUnion(n);
        }
        if (!unionsLookOK(sTree, sUnions)) {
            return false;
        }
        for (int j = 0; j < gTrees.size(); j++) {
            TreeInterface gTree = gTrees.get(j);
            BitUnion [] gUnions = new BitUnion[gTree.getNodeCount()];
            for (int i = 0; i < gTree.getNodeCount(); i++) {
                gUnions[i] = unionArrays.gNodeUnion(j, i);
            }
            if (!unionsLookOK(gTree, gUnions)) {
                return false;
            }
        }
        return true;
    }


    private static boolean unionsLookOK(TreeInterface tree, BitUnion [] unions) {
        for (int n = 0; n<tree.getNodeCount(); n++) {
            Node node = tree.getNode(n);
            if (node.isLeaf()) {
                BitUnion hereU = unions[node.getNr()];
                if (hereU.debugNumberBitsSets() != 1) {
                    return false;
                }
            } else {
                BitUnion hereU = unions[node.getNr()];
                BitUnion lftU  = unions[node.getLeft().getNr()];
                BitUnion rgtU  = unions[node.getRight().getNr()];
                int size = hereU.debugSize();
                for (int b = 0; b < size; b++) {
                    boolean here = hereU.debugBitIsSet(b);
                    boolean lft = lftU.debugBitIsSet(b);
                    boolean rgt = rgtU.debugBitIsSet(b);
                    if ((lft | rgt) != here) {
                        return false;
                    }
                }
            }
            if (node.isRoot()) {
                BitUnion hereU = unions[node.getNr()];
                if (hereU.debugNumberBitsSets() != hereU.debugSize()) {
                    return false;
                }
            }
        }
        return true;
    }


}
