package denim.util;

/**
 * Created by Work on 27/10/2016.
 */
import beast.core.Description;
import beast.core.parameter.BooleanParameter;
import beast.evolution.tree.Node;
        import beast.evolution.tree.Tree;
        import beast.evolution.tree.TreeInterface;
import beast.util.Randomizer;
import denim.GtreeWithEmbedding;


import java.util.*;



@Description("Miscellaneous methods for STEPMILS: Tree and Node utils, debug output.")
public class Misc {


    public static Node siblingOf(Node xN) {
        assert !xN.isRoot();
        Node ancN = xN.getParent();
        if (ancN.getLeft() == xN) {
            return ancN.getRight();
        } else {
            return ancN.getLeft();
        }
    }


    public static Node ancestorBranchAt(Node x, double height) {
        while (!x.isRoot() && x.getParent().getHeight() < height) {
            x = x.getParent();
        }
        return x;
    }


    public static Node mrca(Node x, Node y) {
        while (x != y) {
            if (x.getHeight() < y.getHeight()) {
                x = x.getParent();
            } else {
                y = y.getParent();
            }
        }
        return x;
    }


    // TODO ordered children better than two functions
    public static Node lowestChild(Node x, int [] lowestL) {
        assert !x.isLeaf();
        Node lftN = x.getLeft();
        Node rgtN = x.getRight();
        assert lowestL[lftN.getNr()] != lowestL[rgtN.getNr()];
        return lowestL[lftN.getNr()] < lowestL[rgtN.getNr()] ? lftN : rgtN;
    }


    public static Node highestChild(Node x, int [] lowestL) {
        assert !x.isLeaf();
        Node lftN = x.getLeft();
        Node rgtN = x.getRight();
        assert lowestL[lftN.getNr()] != lowestL[rgtN.getNr()];
        return lowestL[lftN.getNr()] < lowestL[rgtN.getNr()] ? rgtN : lftN;
    }

    public static Node [] orderedChildren(Node x, int [] lowestL) {
        assert !x.isLeaf();
        Node lftN = x.getLeft();
        Node rgtN = x.getRight();
        assert lowestL[lftN.getNr()] != lowestL[rgtN.getNr()];
        Node [] nodes = new Node[2];
        if (lowestL[lftN.getNr()] < lowestL[rgtN.getNr()] ) {
            nodes[0] = lftN;
            nodes[1] = rgtN;
        } else {
            nodes[0] = rgtN;
            nodes[1] = lftN;
        }
        return nodes;
    }



    public static int findInternalNodeIndex(Tree tree, int Nr) {
        // Node.getNr() is documented:
        /**
         * @return number uniquely identifying the node in the tree.
         *         This is a number between 0 and the total number of nodes in the tree
         *         Leaf nodes are number 0 to #leaf nodes -1
         *         Internal nodes are numbered  #leaf nodes  up to #nodes-1
         *         The root node is always numbered #nodes-1
         */
        assert Nr >= tree.getLeafNodeCount();
        assert Nr < tree.getNodeCount();
        return Nr - tree.getLeafNodeCount();
    }




    public static double doBitFlips(BooleanParameter bits) {
        double logHR = 0.0;
        double p = Randomizer.nextDouble();
        for (int i = 0; i < bits.getDimension(); i++) {
            if (Randomizer.nextDouble() > p) {
                bits.setValue(i, !bits.getValue(i));
            }
        }
        return logHR;
    }





    public static String allTreesAsText(TreeInterface sTree, List<Tree> gTrees) {
        StringBuffer buffer = new StringBuffer();
        buffer.append("***************************   All trees   ********************************\n");
        Bindings bindings = Bindings.initialise(sTree, gTrees);
        UnionArrays unionArrays = UnionArrays.initialise(sTree, gTrees, bindings);
        unionArrays.update();

        BitUnion [] sUnions = new BitUnion[sTree.getNodeCount()];
        for (int n = 0; n <sTree.getNodeCount(); n++) {
            sUnions[n] = unionArrays.sNodeUnion(n);
        }
        buffer.append(beastSubtreeAsTextWithHeader(sTree.getRoot(), sUnions));
        for (int j = 0; j < gTrees.size(); j++) {
            Tree gTree = gTrees.get(j);
            BitUnion [] gUnions = new BitUnion[gTree.getNodeCount()];
            for (int i = 0; i < gTree.getNodeCount(); i++) {
                gUnions[i] = unionArrays.gNodeUnion(j, i);
            }
            buffer.append("\nGene tree" + j + "\n");
            buffer.append(beastSubtreeAsTextWithHeader(gTree.getRoot(), gUnions));
        }
        return buffer.toString();
    }



    public static String gToSmappingAsText(int [] gTreeNodeNrToSTreeNodeNr) {
        StringBuffer buffer = new StringBuffer();
        for (int i = 0; i < gTreeNodeNrToSTreeNodeNr.length; i++) {
            buffer.append(i + " " + gTreeNodeNrToSTreeNodeNr[i] + ",  ");
            if ((i+1) % 10 == 0) {
                buffer.append("\n");
            }
        }
        return buffer.toString();
    }


    public static String embeddingParamsAsText(List<GtreeWithEmbedding> gTreesWE) {
        StringBuffer buffer = new StringBuffer();
        for (int j = 0; j < gTreesWE.size(); j++) {
            buffer.append("gTree " + j + ". Mig points, dests\n");
            for (int i = 0; i < gTreesWE.get(j).getTree().getInternalNodeCount(); i++) {
                buffer.append(i + ": " + nonnegIn8Chars(gTreesWE.get(j).getMigPoint(i)) + " " +
                        nonnegIn8Chars(gTreesWE.get(j).getDestinationChoice(i)) + ", ");
                if ((i+1) % 10 == 0) {
                    buffer.append("\n");
                }
            }
            buffer.append("\n");
        }
        return buffer.toString();
    }


    public static String CountsIntensitiesAsText(int [][] coalCounts,  int [][] migEventsCounts,
                                                 int [][] migLeaveCounts, double [][] coalIntensities) {
        StringBuffer buffer = new StringBuffer();
        buffer.append("coalCounts\n");
        for (int n = 0; n < coalCounts.length; n++) {
            buffer.append("sTree node " + n + " locus ");
            for (int j =0; j < coalCounts[n].length; j++) {
                buffer.append(j + " " + coalCounts[n][j] + ", ");
            }
            buffer.append("\n");
        }

        buffer.append("migEventsCounts\n");
        for (int n = 0; n < migEventsCounts.length; n++) {
            buffer.append("sTree node " + n + " locus ");
            for (int j =0; j < migEventsCounts[n].length; j++) {
                buffer.append(j + " " + migEventsCounts[n][j] + ", ");
            }
            buffer.append("\n");
        }

        buffer.append("migLeaveCounts\n");
        for (int n = 0; n < migLeaveCounts.length; n++) {
            buffer.append("sTree node " + n + " locus ");
            for (int j =0; j < migLeaveCounts[n].length; j++) {
                buffer.append(j + " " + migLeaveCounts[n][j] + ", ");
            }
            buffer.append("\n");
        }

        buffer.append("coalIntensities\n");
        for (int n = 0; n < coalIntensities.length; n++) {
            buffer.append("sTree node " + n + " locus ");
            for (int j =0; j < coalIntensities[n].length; j++) {
                buffer.append(j + " " + coalIntensities[n][j] + ", ");
            }
            buffer.append("\n");
        }


        return buffer.toString();
    }


    public static String beastSubtreeAsNewick(TreeInterface tree, Node node) {
        ArrayList<String> nodeLabels = new ArrayList<>(tree.getNodeCount());
        for (int n = 0; n < tree.getNodeCount(); n++) {
            nodeLabels.add(tree.getNode(n).getID() == null ? "" + n : tree.getNode(n).getID());
        }
        return node.toString(nodeLabels);
    }


    public static String beastSubtreeAsTextWithHeader(Node node, BitUnion [] unions) {
        String header = "topology                                 idx height\n";
        String s = "";
        Stack<Integer> x = new Stack<Integer>();
        return header + beastSubtreeAsText(node, s, x, 0, "", unions);
    }


    private static String beastSubtreeAsText(Node node, String s, Stack<Integer> x, int depth, String b, BitUnion [] unions) {
        Integer[] y = x.toArray(new Integer[x.size()]);
        StringBuffer indent = new StringBuffer();
        for (int i = 0; i < depth; i++) {
            indent.append("  ");
        }
        for (int i = 0; i < y.length; i++) {
            indent.replace(2 * y[i], 2 * y[i] + 1, "|");
        }
        if (b.length() > 0) {
            indent.replace(indent.length() - b.length(), indent.length(), b);
        }
        s += indent;
        s += beastNodeAsText(node, indent.length(), unions);
        s += System.getProperty("line.separator");
        String subs = "";
        if (!node.isLeaf()) {
            x.push(depth);
            subs += beastSubtreeAsText(node.getLeft(), "", x, depth + 1, "-", unions);
            x.pop();
            subs += beastSubtreeAsText(node.getRight(), "", x, depth + 1, "`-", unions);
        }
        return s + subs;
    }


    private static String beastNodeAsText(Node node, int indentlen, BitUnion [] unions) {
        StringBuilder s = new StringBuilder();
        Formatter formatter = new Formatter(s, Locale.US);
        if (node.isLeaf()) {
            formatter.format("%s ", node.getID());
        } else {
            formatter.format("%s ", "+");
        }
        while (s.length() < 40 - indentlen) {
            formatter.format("%s", " ");
        }
        int n = node.getNr();
        formatter.format("%3d ", n);
        formatter.format("%s ", nonnegIn8Chars(node.getHeight()));
        formatter.format("%30s", unions[node.getNr()].asText());
        return s.toString();
    }


    public static String nonnegIn8Chars(double x) {
        StringBuilder s = new StringBuilder();
        Formatter formatter = new Formatter(s, Locale.US);
        if (x < 0) {
            formatter.format("%8s", "NA");
        } else if (x == 0.0) {
            formatter.format("%8s", "zero");
        } else if (x < 1e-3) {
            formatter.format("%8.2e", x);
        } else if (x < 9.999) {
            formatter.format("%8.5f", x);
        } else if (x < 99.99) {
            formatter.format("%8.4f", x);
        } else if (x < 999.9) {
            formatter.format("%8.3f", x);
        } else if (x < 9999) {
            formatter.format("%8.2f", x);
        } else {
            formatter.format("%8.0f", x);
        }
        return s.toString();
    }

}